#  Copyright 2014 Klaudiusz Staniek
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
 fiblary.v3.base
 ~~~~~~~~~~~~~~~~~

 Home Center Controller Base Class Implementation

"""

import fiblary3.external.jsonpath as jsonpath

from itertools import filterfalse
import logging
import six

from fiblary3.common import exceptions
from fiblary3.common.utils import quote_if_string


_logger = logging.getLogger(__name__)


def _check_items(obj, searches):
    def _check_properties(attr, value):
        properties = getattr(obj, 'properties', None)
        if properties:
            return properties.get(attr) == value
        return False
    found = all(((getattr(obj, attr, None) == value) or
                _check_properties(attr, value)) for (attr, value) in searches)

    return found


class MinimalController(object):
    """Minimal controller class used as a base class for more
    specific controllers.

    It implements simple get method for read only resources.
    """

    RESOURCE = ''
    API_PARAMS = ()
    """API PARAMS can be directly passed to the RESTApi
    This speed up the list operation
    """

    def __init__(self, http_client, model):
        self.http_client = http_client
        self.model = model

    def _get(self, **kwargs):

        try:
            item = self.http_client.get(self.RESOURCE, params=kwargs).json()
        except exceptions.ConnectionError:
            return None
        except exceptions.HTTPNotFound:
            return None

        return item

    def get(self):
        """Returns :class:`models.Model` object representing. This is
        applicable for single instance requests i.e. login, info, weather

        :returns: :class:`models.Model` object
        """
        item = self._get()
        return self.model(item)


class ReadOnlyController(MinimalController):
    """Read only controller class used for controllers using get and list
    methods only
    """
    JSON_CONDITION_BASE = "(@.{0}=={1})"

    def get(self, item_id):
        """Returns :class:`models.Model` object representing an item
        identified by ``item_id``

        :param item_id: This is an id of the requested object. The item_id
        encoded in GET request as the 'id' parameter
        :returns: :class:`models.Model` object
        """

        # item_it must not be None.
        if item_id is None:
            _logger.debug("Called 'get' method with none id")
            return None

        params = {"id": item_id}
        item = self._get(**params)
        return self.model(item)

    def list(self, **kwargs):
        """
        :param kwargs: This is a dictionary of parameters passed to GET
        :type kwargs:
        :return: Returns an iterator
        :rtype: iterator function
        """
        _logger.debug(self.RESOURCE)
        _logger.debug(type(self))
        _logger.debug(kwargs)

        # Pop jsonpath if exists and pass the rest of arguments to API
        # for some API calls home center handles additional parameters

        json_path = kwargs.pop('jsonpath', None)

        # Home center ignores unknown parameters so there is no need to
        # remove them from REST request.
        items = self.http_client.get(self.RESOURCE, params=kwargs).json()

        # if there is no explicit defined json_path parameters
        if json_path is None:
            for value in self.API_PARAMS:
                kwargs.pop(value, None)

            condition_expression = ""
            for k, v in six.iteritems(kwargs):
                if k.startswith('p_'):  # search for properties
                    k = "properties." + k[2:]
                condition_expression += self.JSON_CONDITION_BASE.format(
                    k,
                    quote_if_string(v))
                condition_expression += " and "

            if condition_expression is not "":
                # filter the results with json implicit built from
                # remaining parameters

                json_path = "$[?({})]".format(condition_expression[:-5])
            _logger.debug("Implicit JSON Path: {}".format(json_path))

        if json_path:
            _logger.debug("JSON Path: {}".format(json_path))
            filtered_items = jsonpath.jsonpath(items, json_path)
            if filtered_items:
                items = filtered_items
            else:
                items = []

        # in case there is only one item
        items = items if isinstance(items, list) else [items]
        return filterfalse(lambda i: i is None, map(self.model, items))

    def find(self, **kwargs):
        """Find single item with attributes matching ``**kwargs``.
        It also handles nested properties as a keywords.
        """
        num_matches = 0
        found = None
        items = self.list(**kwargs)
        for item in items:
            found = item
            num_matches += 1
            if num_matches > 1:  # raise exception not waiting
                                    # for the whole list walkthrough
                    raise exceptions.NoUniqueMatch

        if not found:
            msg = "No matching {0}".format(kwargs)
            raise exceptions.NotFound(msg)

        return found

    # depreciated
    def findall(self, **kwargs):
        _logger.warn(
            "The 'findall' method was depreciated. Use 'list' instead")
        return [item for item in self.list(**kwargs) if _check_items(
            item,
            six.iteritems(kwargs))]


class CommonController(ReadOnlyController):
    """Common controller class used as a base class for controllers
    implementing method changing the data
    """

    def create(self, **kwargs):
        item = self.model(kwargs)
        for (key, value) in list(kwargs.items()):
            setattr(item, key, value)
        try:
            response = self.http_client.post(self.RESOURCE, data=item)
        except exceptions.ConnectionError:
            return None

        try:
            item = response.json()
        except ValueError:
            _logger.warning(
                "Invalid JSON format. Received: '{}'".format(response.text))
            return None

        return self.model(item)

    def delete(self, item_id):
        url = '{0}?id={1}'.format(self.RESOURCE, item_id)
        self.http_client.delete(url)
        return

    def update(self, data):
        try:
            response = self.http_client.put(self.RESOURCE, json=data)
        except exceptions.ConnectionError:
            return None

        try:
            item = response.json()
        except ValueError:
            _logger.warning(
                "Invalid JSON format. Received: '{}'".format(response.text))
            return None

        return self.model(item)



#  Copyright 2020 Daniel Pervan
#  Copyright 2014 Klaudiusz Staniek
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
 fiblary.variables
 ~~~~~~~~~~~~~~

 Home Center Variable Manager Implementation
"""

import logging

from fiblary3.client.v5 import base
from fiblary3.common import exceptions

# TODO(kstaniek): Handle predefined variables

_logger = logging.getLogger(__name__)


class Controller(base.CommonController):
    RESOURCE = 'globalVariables'

    def get(self, item_id):
        url = '{0}/{1}'.format(self.RESOURCE, item_id)
        item = self.http_client.get(url).json()
        return self.model(item)

    def delete(self, item_id):
        url = '{0}/{1}'.format(self.RESOURCE, item_id)
        self.http_client.delete(url)
        return

    def set(self, name, value):
        data = {
            'value': value
        }
        try:
            url = "%s/%s" % (self.RESOURCE, name)
            item = self.http_client.put(url, json=data)

            _logger.info('Variable set: {0}="{1}"'.format(name, value))
        except exceptions.HTTPNotFound:
            _logger.info('Variable created: {0}="{1}"'.format(name, value))
            item = self.create(data)

        return self.model(item.json())



#  Copyright 2014 Klaudiusz Staniek
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
 fiblary.common.event
 ~~~~~~~~~~~~~~~~~~~~

 Event Implementation
"""
import logging
import threading

from fiblary3.common import exceptions

try:
    import queue as queue
except Exception:
    import queue

import inspect

_logger = logging.getLogger(__name__)


class EventQueue(threading.Thread):

    def __init__(self, name=None):
        super(EventQueue, self).__init__(name=name or self.__class__.__name__)
        self.queue = queue.Queue()
        self._stopper = threading.Event()

        self.n = 0
        self.serving = None
        self.daemon = True  # stop unconditionally on exit

        _logger.info("Starting the event queue for {}".format(self.name))
        self.start()

    def error(self, error, function, a=(), kw=None):
        """Called when function raises error.
        """
        _logger.exception(
            "EventQueue event raised exception ({}, {}, {}):{}".format(
                function, a, kw or {}, error)
        )

    def put(self, event, function=None, a=(), kw=None):
        """Queue the call to function, return the number of the call.
        """
        self.queue.put((event, function, a, kw or {}))
        return self.n + self.queue.qsize()

    def stop(self):
        _logger.info("Stopping the event queue for {}".format(self.name))
        self._stopper.set()
        self.put("EXIT")

    def run(self):
        """A blocking loop which continually calls functions as specified
        in self.queue.
        """
        while not self.stopped():
            try:
                event, function, a, kw = self.queue.get(True, 120)
            except queue.Empty:
#                _logger.warning("Event queue timeout: {}".format(self.name))
                continue

            if event == "EXIT":
                break

            _logger.debug("Calling on {} changed handlers".format(event))
            self.serving = (function, a, kw)
            try:
                function(*a, **kw)
            except Exception as e:
                self.error(e, function, a, kw)
            self.n += 1

    def stopped(self):
        return self._stopper.isSet()


def queue_event(f):
    """Decorator which queues method/function calls in
    self.eventqueue and self.name [if f is a method whose
    first argument is 'self'],
    otherwise f.eventqueue.
    """
    args = inspect.getargspec(f)[0]
    if args and (args[0] == 'self'):
        def decorated(self, *a, **kw):
            if self.event_queue:
                self.event_queue.put(self.__repr__(), f, (self,) + a, kw)
    else:
        f.event_queue = EventQueue()

        def decorated(*a, **kw):
            f.event_queue.put(f.__name__, f, a, kw)
    decorated.__name__ = f.__name__
    decorated.__doc__ = f.__doc__
    return decorated


class EventHook(object):
    """EventHook as suggested from Michael Foord in his Event Pattern:
    http://www.voidspace.org.uk/python/weblog/arch_d7_2007_02_03.shtml#e616

    Extended with queue event
    """

    def __init__(self, name=""):
        self.__handlers = []
        self.name = name
        self.event_queue = None

    def get_handler_count(self):
        return len(self.__handlers)

    def __iadd__(self, handler):
        if not self.event_queue:
            self.event_queue = EventQueue(self.name)
        self.__handlers.append(handler)
        return self

    def __isub__(self, handler):
        try:
            self.__handlers.remove(handler)
        except ValueError:
            raise exceptions.HandlerNotFound

        if self.get_handler_count() == 0:
            self.event_queue.stop()
            self.event_queue.join()
            del self.event_queue
            self.event_queue = None
        return self

    def __repr__(self):
        return self.name

    @queue_event
    def __fire(self, *args, **keywargs):
        for handler in self.__handlers:
            handler(*args, **keywargs)

    __call__ = __fire



#  Copyright 2014 Klaudiusz Staniek
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
 fiblary.common.utils
 ~~~~~~~~~~~~~~~~~~~~

 Common Utilities Implementation
"""


import sys


def import_module(import_str):
    """Import a module."""
    __import__(import_str)
    return sys.modules[import_str]


def import_versioned_module(module, version, submodule=None):
    module = 'fiblary3.%s.%s' % (module, version)
    if submodule:
        module = '.'.join((module, submodule))
    return import_module(module)


def quote_if_string(value):
    if type(value) is str:
        return "'{0}'".format(value)
    else:
        return value



#  Copyright 2014 Klaudiusz Staniek
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
 fiblary.common.beacon
 ~~~~~~~~~~~~~~~~~~~~~

 Home Center Beacon Implementation

"""


from fiblary3.common.net import get_mac
import socket
import threading

SERVER_TIMEOUT = 0.2


class Beacon(threading.Thread):
    def __init__(self, serial):
        threading.Thread.__init__(self)
        self.port = 44444
        self.key = 'FIBARO'
        self.quit = False
        self.serial = serial

        self.mac = get_mac()

    def run(self):
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.settimeout(SERVER_TIMEOUT)
        sock.bind(("", self.port))

        while not self.quit:
            try:
                message, address = sock.recvfrom(len(self.key))
            except socket.error:
                continue

            if message != self.key:
                # not my client
                continue

            sock.sendto('ACK {} {}'.format(self.serial, self.mac), address)

        sock.close()

# b = Beacon('HC2-999329')
# b.daemon = True
# b.start()
# time.sleep(10000)



#  Copyright 2020 Daniel Pervan
#  Copyright 2014 Klaudiusz Staniek
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
 fiblary.devices
 ~~~~~~~~~~~~~~

 Home Center Device Manager Implementation
"""

from fiblary3.client.v5 import base
from fiblary3.common import exceptions

import logging


_logger = logging.getLogger(__name__)


class Controller(base.CommonController):
    RESOURCE = 'devices'
    API_PARAMS = ('id', 'type', 'roomID')

    def action(self, device_id, action, *args):
        cmd = 'devices/{}/action/{}'.format(device_id, action)
        if args:
            data = '{ "args":'
            for i, arg in enumerate(args):
                if (i > 1):
                    data = '{}, '.format(data)
                data = '{}"{}"'.format(data, arg)
        else:
            data = "{}"
        resp = self.http_client.post(cmd, data=data)
        if resp.status_code != 200 and resp.status_code != 202:
            exceptions.from_response(resp)

    def update(self, data):
        try:
            """
            HC2 does not like those properties to be PUT
            """
            data.properties.pop("associationView")
            data.properties.pop("associationSet")

        except Exception:
            pass

        return super(Controller, self).update(data)



#  Copyright 2014 Klaudiusz Staniek
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
 fiblary.devices
 ~~~~~~~~~~~~~~

 Home Center Device Manager Implementation
"""

from fiblary3.client.v3 import base
from fiblary3.common import exceptions

import logging


_logger = logging.getLogger(__name__)


# TODO(kstaniek):
# http://hc2/api/devices?type=satel_partition
# http://hc2/api/callAction?deviceID=312&name=removeFailedNodeId
# http://hc2/api/callAction?deviceID=38&name=wakeUpDeadDevice


class Controller(base.CommonController):
    RESOURCE = 'devices'
    API_PARAMS = ('id', 'type', 'roomID')

    def action(self, device_id, action, *args):
        cmd = "callAction?deviceID={0}&name={1}".format(device_id, action)
        for i, arg in enumerate(args, 1):
            cmd = '{0}&arg{1}={2}'.format(cmd, i, arg)

        resp = self.http_client.get(cmd)
        if resp.status_code != 200 and resp.status_code != 202:
            exceptions.from_response(resp)

    def update(self, data):
        try:
            """
            HC2 does not like those properties to be PUT
            """
            data.properties.pop("associationView")
            data.properties.pop("associationSet")

        except Exception:
            pass

        return super(Controller, self).update(data)
